/***************************** LICENSE START ***********************************

 Copyright 2012 ECMWF and INPE. This software is distributed under the terms
 of the Apache License version 2.0. In applying this license, ECMWF does not
 waive the privileges and immunities granted to it by virtue of its status as
 an Intergovernmental Organization or submit itself to any jurisdiction.

 ***************************** LICENSE END *************************************/

#include "CptecAccess.h"

//#include <string.h>

void CptecAccess::serve(MvRequest& in,MvRequest& out)
{
cout << "Entry CPTEC Access::serve..." << endl;
in.print();

   int i;

   // Build the path
   string spath;
   if ( !buildPath(in,spath) )
      return;

   // Build the filename(s)
   vector <string> vfiles;
   if ( !buildFilenames(in,vfiles) )
      return;

   // Get the requested PARAMs
   const char* caux;
   int nparams = in.iterInit( "PARAMETER" );
   map <string, int> params;
   string param;
   for ( i = 0; i < nparams; ++i )
   {
      in.iterGetNextValue( caux );
      param = caux;
      params[param] = 1;
   }

   // Get the requested LEVELs
   int nlevels = in.iterInit( "LEVEL_LIST" );
   map <long, int> levels;
   double dlevel;
   for ( i = 0; i < nlevels; ++i )
   {
     in.iterGetNextValue( dlevel );
     levels[(long)dlevel] = 1;
   }
	
   // Build the requested fieldset
   // For each input file filter the requested fields by PARAM/LEVEL
   // and add them to the output GRIB file

   // Auxilliary variables for GribApi
   char shortName[20];
   size_t len;
   int error=0;	
   grib_handle* h=NULL;
   grib_context* c=grib_context_get_default();

   // Create the output file name
   string outname(marstmp());
   for ( i = 0; i < vfiles.size(); ++i )
   {
        // Full filename
        string ffile = spath + vfiles[i];

        // Open the input file
        FILE* f = fopen(ffile.c_str(),"r");
        if ( !f )
        {
             string error = "FILE NOT FOUND: ";
             error += ffile.c_str();
             setError(0, error.c_str());
             return;
        }

        // Loop on all GRIB messages in file
        long level;
        while ((h=grib_handle_new_from_file(c,f,&error))!=NULL)
        {
            len=20;
            grib_get_string(h,"shortName",shortName,&len);
            param = string(shortName);
            grib_get_long(h,"level",&level);
            if (params.find(param) != params.end())
            {
                if (levels.find(level) != levels.end())
                {
                   cout << "Found PARAM " << param << " LEVEL " << level << " in " <<  vfiles[i] << endl;
                   grib_write_message(h,outname.c_str(),"a");
                }
            }

            grib_handle_delete(h);
        }

        fclose(f);
   }

   // Create output request
   out.setVerb("GRIB");
   out("PATH") = outname.c_str();
out.print();

   return;
}

bool CptecAccess::buildPath ( MvRequest& in, string& spath )
{
   // Build the path
   // 1. Date/time related string
   const char* date = in("DATE");
   const char* time = in("TIME");
   string yyyymm(date,6);
   string dd(date,6,2);
   string hh(time,2);
   string sdate = yyyymm + '/' + dd + hh;

//eta 20
// /rede/eva01/modoper/tempo/regional/etap_20km/grbctl/2012/Jan/11

//eta 15 - 1h
// /rede/contg02/stornext/oper/tempo/Eta/ams_15km/temp/201201/1100/grib

   // 2. Fixed part of the path. Try to get it from the 
   // environment variable: METVIEW_CPTEC_ACCESS_PATH. 
   // If not defined, use a default value.
   const char* fpath = getenv("METVIEW_CPTEC_ACCESS_PATH");
//   spath = fpath ? fpath : "/rede/tupa_tempo/";
   spath = fpath ? fpath : "/mnt/";

   // 3. Part of the path related to the model
   const char* model = in("MODEL");
   if ( strcmp(model,"ETA153") == 0 )
     spath += "Eta/ams_15km/" + sdate + "/grbctl/";
   else if ( strcmp(model,"ETA151") == 0 )
//     spath += "Eta/ams_15km/temp/" + sdate + "/grib/";
     spath += "eta15/stornext/oper/tempo/Eta/ams_15km/temp/" + sdate + "/grib/";
   else if ( strcmp(model,"ETA20") == 0 )
     spath += "eta20/modoper/tempo/regional/etap_20km/grbctl/";
   else if ( strcmp(model,"GLOBAL") == 0 )
     spath += "/";
   else
   {
        string error = "ERROR: MODEL not recognized ";
        setError(1, error.c_str());
        return false;
   }

   return true;
}

bool CptecAccess::buildFilenames ( MvRequest& in, vector <string>& vfn )
{
   int i,j,k;

   // Part of the filename related to the model
   string smodel;
   const char* model = in("MODEL");
   if ( strcmp(model,"ETA153") == 0 || strcmp(model,"ETA151") == 0 )
      smodel = "eta_15km_";
   else if ( strcmp(model,"ETA20") == 0 )
      smodel = "";
   else if ( strcmp(model,"GLOBAL") == 0 )
      smodel = "ll20gfsa";
   else
   {
        string error = "ERROR: MODEL not recognized ";
        setError(1, error.c_str());
        return false;
   }

   // Get all values related to parameters DATE/TIME/STEP
   double xx;
   int ndates = in.iterInit( "DATE" );
   vector <double> dates;
   dates.reserve(ndates);
   for ( i = 0; i < ndates; ++i )
   {
      in.iterGetNextValue( xx );
      dates.push_back(xx);
   }

   int ntimes = in.iterInit( "TIME" );
   vector <double> times;
   times.reserve(ntimes);
   for ( i = 0; i < ntimes; ++i )
   {
      in.iterGetNextValue( xx );
      times.push_back(xx/100.); //hhmm
   }

   int nsteps = in.iterInit( "STEP" );
   vector <double> steps;
   steps.reserve(nsteps);
   for ( i = 0; i < nsteps; ++i )
   {
      in.iterGetNextValue( xx );
      steps.push_back(xx);
   }

   // Build all filenames
   char buf[12];
   for ( i = 0; i < ndates; i++ )
   {
      MvDate dd(dates[i]);
      for ( j = 0; j < ntimes; j++ )
      {
         // Analise time stamp string
	 MvDate danal = dd + times[j]/24.;
	 danal.Format("yyyymmddHH",buf);
         string cs = smodel + buf + '+';

         // Forecast time stamp string
         for ( k = 0; k < nsteps; k++ )
         {
            // Forecast date
	    MvDate dfore = danal + steps[k]/24.;
	    dfore.Format("yyyymmddHH",buf);
            vfn.push_back(cs + buf + ".grb");
	 }
      }
   }

   return true;
}

int main(int argc,char **argv)
{
	MvApplication theApp(argc,argv);
	CptecAccess data("CPTEC_ACCESS");

	theApp.run();
}
