/***************************** LICENSE START ***********************************

 Copyright 2012 ECMWF and INPE. This software is distributed under the terms
 of the Apache License version 2.0. In applying this license, ECMWF does not
 waive the privileges and immunities granted to it by virtue of its status as
 an Intergovernmental Organization or submit itself to any jurisdiction.

 ***************************** LICENSE END *************************************/

//#include "inc_iostream.h"
//#include "MvException.h"

//#include <math.h>

//#include "Metview.h"
//#include "MvTephiFrame.h"
//#include "MvTephiParams.h"
//#include "Cached.h"
//#include "MvObs.h"
#include "GribThermo.h"
#include "DataThermo.h"
#include "MvException.h"
#include "MvNetCDF.h"

//static Cached GRIB("GRIB");
//static Cached BUFR("BUFR");

GribThermo::GribThermo() :
                   Thermo("GRIB_THERMO")
{
   //type_ = "MXSECTION";
   //view_ = "MXSECTIONVIEW";
}

GribThermo::GribThermo(const char*kw) :
                   Thermo(kw)
{
   //type_ = kw;
   //view_ = "MXSECTIONVIEW";
}

GribThermo::~GribThermo()
{
}

// Retrieve parameters from the input request and make a consistency check.
// There are 4 types of input request:
// 1. DATA_DROPPED: Grib dropped into a View: MXSECT, DATA(), _CONTEXT(),...
// 2. DATA_MODULE_DROPPED: Data module dropped into a View: MXSECT, DATA(),...
// 3. DATA_MODULE: Process a data module: MXSECT, ...
// 4. netCDF dropped in a View:
//
// Actions for each 4 types:
// 1. Use the original View.
// 2. Use the original View and check if the input XSectData parameters are
//    consistent with the View.
// 3. Build a new View.
// 4. Use the original View and check if the netCDF parameters are consistent
//    with the View.
//
bool GribThermo::getInputParameters( MvRequest& in )
{
   // MAYBE REMOVE IT LATER
   // Get input parameters
   if( !pars_.init(in) )
   {
      setError( 1, pars_.errorMsg_.c_str() );
      return false;
   }

#if 0
   // Retrieve original View if exists, e.g. if a data or a
   // xs/av/vpdata module was dropped into a view
   MvRequest viewRequest;
   if ( (const char*)in("_CONTEXT") )
      viewRequest = in("_CONTEXT");

   // Retrieve xsectdata parameters if exist, e.g. if a xs/av/vpata module
   // was dropped into a view or if it was called to be processed
   const char* verb = (const char*)in("_VERB");
   bool moduleDropped = this->isDataModuleDropped(verb);

   // Check the type of input request to be processed
   MvRequest commonRequest;
   int procType;
   if ( viewRequest )
      procType = moduleDropped ? XS_DATA_MODULE_DROPPED : XS_DATA_DROPPED;
   else
      procType = XS_DATA_MODULE;

   // Type of input request is a DATA_DROPPED into a View
   // Get information from the View request
   double pBottom=0., pTop=0.;
   if ( procType == XS_DATA_DROPPED )
   {
      // Retrieve 'original' request from the View
      commonRequest = getAppView( viewRequest );

      // Retrieve parameters
      if ( !getAppParameters(commonRequest, appInfo) )
         return false;
   }
   // Process a data module
   // Get information from the module icon parameters
   else if ( procType == XS_DATA_MODULE )
   {
      // Retrieve parameters
      if ( !this->getAppParameters(in, appInfo) )
         return false;
   }
   // Process a data module dropped into a view
   // Get information from the module and view. Check if they are compatible.
   else if ( procType == XS_DATA_MODULE_DROPPED )
   {
      // Retrieve 'original' request from the View
      commonRequest = getAppView( viewRequest );

      // Retrieve 'original' request from the data module
      MvRequest modRequest = data.getSubrequest("_ORIGINAL_REQUEST");
      if ( !modRequest )
         modRequest = in;

      // Consistency check, only for non-default View request
      if ( !((const char*)commonRequest("_DEFAULT") && (int)commonRequest("_DEFAULT") == 1) )
         if ( !this->consistencyCheck(commonRequest, modRequest) )
            return false;

      // Retrieve application specific parametes
      if ( !getAppParameters( modRequest, appInfo ) )
         return false;
   }


   appInfo.processType(procType);
#endif
   return true;
}

bool GribThermo::processData()
{
   // If it is not a GRIB data then it has been already processed
   if ( strcmp(pars_.dataReq_.getVerb(),"GRIB") != 0 )
      return true;

   // Create a temporary netCDF file
   string ncFileName = this->ncFileName();
   MvNetCDF netcdf(ncFileName,'w');
   if ( !netcdf.isValid() )
   {
      setError(1,"ERROR: Can not open a new netCDF file");
      return false;
   }

   // Initialize fildset iterator.
   // The order of the sorting is very important. Parameters STEP, TIME and
   // DATE should be the slowest indexes to be sorted since they define a new
   // plot. The algorithm relies that all fields related to the same Level id
   // are next to each other.
   MvFieldSet fs(pars_.dataReq_);
   MvFieldSetIterator iter(fs);
   iter.sort("LEVELIST");
   iter.sort("STEP");
   iter.sort("TIME");
   iter.sort("DATE");

   // Main loop: process fieldset
   MvField field = iter();
   while( field )   //-- loops dates, i.e. thermoD plots
   {
      // Create data strucutre according to the data representation
      MvThermoDataBase* tep = MvThermoDataFactory( field, pars_ );
      if( !tep )
      {
         setError( 13, "Thermo: unimplemented input field level type!" );
         return false;
      }

      try
      {
         // Read and compute all fields related to one plot
         tep->readOneSetOfFields( iter );

         // Compute final values and save results to the netCDF file
         tep->generate( netcdf );
      }
      catch( MvException& e )
      {
         cerr << "Exception caught: " << e.what() << endl;
         setError( 13, e.what() );
         return false;
      }

      // Check if wind info is available
      if ( tep->windExist() )
         pars_.windFlag(true);

      // Get first field related to the next plot
      field = tep->returnToken();

      delete tep;
   }

   // Close netCDF file
   netcdf.close();

   return true;
#if 0
   MvRequest cdfreq("NETCDF");
   cdfreq("PATH") = netcdfName.c_str();
   out = out + cdfreq;
   return true;

      // It is a netCDF data and it has been already processed.
      // Build the output request
      //out = this->createOutputRequest( in );
/////
   // Initialize fildset iterator
   MvFieldSet fs(data);
   MvFieldSetIterator iter(fs);
   MvField field = iter();

   // Retrieve initial info
   this->getInitialFieldInfo(field,appInfo,north,south,east,west);

   // First loop: build field keystring and get the Level info.
   // The LNSP matrix data is also computed, but the other matrices
   // are computed in the second loop.
   iter.rewind();
   while( field = iter() )
   {
      // Retrieve info from the field request
      MvRequest rq=field.getRequest();
      iparam = rq("PARAM");
      const char *lev = rq("LEVELIST");
      istep  = rq("STEP");

      // Check that this field's vital statistics match those of the others
      // - this is important only in the Average application
      if ( !fieldConsistencyCheck(field,north,south,east,west) )
         return false;

      // Calculate the data matrix for LNSP field
      generateKey(keystr,iparam,idate,itime,istep,expver);
      if ( iparam == LnPress )
      {
         fillValues( appInfo,field,splin );
         ParamInfo* par = new ParamInfo();
         params.insert(ParamPair(keystr,par) );
      }
      else
      {
         if ( iparam == U_FIELD )
            foundU = true;
         else if ( iparam == V_FIELD )

         ParamInfo* par = new ParamInfo();

         ParamInsertPair inserted = params.insert(ParamPair(keystr,par));
         ParamInfo *pp = (*(inserted.first)).second;
      }
   } //end while (first loop)

   // Check if fieldset is compatible with the application
   if ( this->consistencyCheck( appInfo, params ) == false )
   {
      setError(1,"ERR"); return false;
   }

   // Create a temporary netCDF file
   ...
   // Write netCDF global attributes
   if ( !this->writeGlobalAttributesNetcdf(netcdf,params,appInfo) )
      return false;

   // Write initial variables to the netCDF file: levels, time,
   // geographical coordinates
   iter.rewind();
   field = iter();
   if ( !this->writeLevelInfoNetcdf(netcdf,appInfo,params,&field,splin) )
      return false;
   if ( !this->writeTimeInfoNetcdf(netcdf,params,ntimes) )
      return false;
   if ( !this->writeGeoCoordsNetcdf( netcdf, appInfo ) )
      return false;

   // Initialize variables
   string lastKey        = "FIRSTFIELD";
   int currentGenerated  = 0, lastNrGenerated = -1;
   MvField lastField;

   // Initialize fieldset iterator
   iter.rewind();
   iter.sort("LEVELIST");

   // Second loop: generate data
   ParamIterator paramIter;
   while( field = iter() )
   {
      currentGenerated++;
      MvRequest r = field.getRequest();
      iparam = r("PARAM");

      // Generate key
      generateKey(keystr,iparam,idate,itime,istep,expver);

      // Calculate the data matrix for non LNSP field
      if ( iparam != LnPress ) // Already done for lnsp
         bool ok = fillValues(appInfo,field,xint);

      // Write out the data if we have the end of what can be grouped together.
      if ( lastKey != keystr && lastKey != string("FIRSTFIELD") )
      {
         if ( !generateData(appInfo,params,netcdf,lastField,lastKey) )
         {setError(1,"");return false; }
         lastNrGenerated = currentGenerated;
      }

      lastKey = keystr;
      lastField = field;
   }  // end while (second loop)

   // Write out the last input field data
   if ( lastNrGenerated <=  currentGenerated )
   {
      if ( !generateData(appInfo,params,netcdf,lastField,lastKey) )
         setError(1,"");return false;
   }

   // Write additional fields, if it is needed
   if ( ! this->generateExtraData(netcdf,appInfo,params,iter) )
      setError(1,"C"); return false;

   // Close netCDF file
   netcdf.close();

   // Get default parameter to be plotted
   ParamInfo* parInfo = plotVariable(params);
   if ( !parInfo )
      return false;

   // Create output request
   out = this->createOutputRequest( appInfo, netcdfName, parInfo );

   // Free memory for parameters
   for (paramIter = params.begin(); paramIter != params.end(); paramIter++)
      delete (*paramIter).second;
#endif
}

bool GribThermo::createOutputRequest(MvRequest& out)
{
   //MUDE ESTE TESTE PARA UMA FUNCAO. SUBSTITUA TAMBEM NA SUBROUTINA ACIMA
   // If it is not a GRIB data then it has been already processed
//   if ( strcmp(pars_.dataReq_.getVerb(),"GRIB") != 0 )
//      return createSimpleOutputRequest(out);

   // Get default parameter to be plotted
   //ParamInfo* parInfo = plotVariable(params);
   //if ( !parInfo )
   //   return false;

   // Create netCDF data request
   // Get action mode
   string actionMode = this->actionMode();
//(actionMode == "analyse" || actionMode == "prepare") )
//verb = "NETCDF";

   MvRequest ncDataReq("NETCDF");
   ncDataReq("PATH") = this->ncFileName().c_str();
   ncDataReq("TEMPORARY") = 1;

   // Create Temperature output request
   MvRequest outd("NETCDF_XY_POINTS");
   outd("NETCDF_DATA") = ncDataReq;
   outd("NETCDF_Y_VARIABLE") = sPRESSURE.c_str();
   outd("NETCDF_X_VARIABLE") = sTEMP.c_str();

   MvRequest gReq("MGRAPH");
   gReq("GRAPH_LINE_COLOUR") = "red";
   gReq("GRAPH_LINE_THICKNESS") = 8;
   outd = outd + gReq;

   // Create Dewpoint output request
   MvRequest outtd("NETCDF_XY_POINTS");
   outtd("NETCDF_DATA") = ncDataReq;
   outtd("NETCDF_Y_VARIABLE") = sPRESSURE.c_str();
   outtd("NETCDF_X_VARIABLE") = sDEWPOINT.c_str();

   gReq("GRAPH_LINE_COLOUR") = "red";
   gReq("GRAPH_LINE_STYLE") = "dash",
   gReq("GRAPH_LINE_THICKNESS") = 8;
   outd = outd + outtd + gReq;

   // Create wind output request
   if ( pars_.windFlag() )
   {
      MvRequest outw("NETCDF_XY_VECTORS");
      outw("NETCDF_DATA") = ncDataReq;
      outw("NETCDF_X_POSITION_VARIABLE")  = sXWIND.c_str();
      outw("NETCDF_Y_POSITION_VARIABLE")  = sPWIND.c_str();
      outw("NETCDF_X_COMPONENT_VARIABLE") = sU.c_str();
      outw("NETCDF_Y_COMPONENT_VARIABLE") = sV.c_str();

      MvRequest wReq("MWIND");
      wReq("WIND_FIELD_TYPE")  = "flags";
      wReq("WIND_FLAG_COLOUR") = "black";

      outd = outd + outw + wReq;
   }
   // Create customised Cartesian View request.
   // If an icon was dropped into a view, uPlot will ignore it.
   // Mode-specific options
   MvRequest viewReq("CARTESIANVIEW");
   viewReq("MAP_PROJECTION") = "tephigram";

   // If action is not visualisation related then return the netcdf data.
   // Also, add the visualisation and original requests as hidden parameters.
   // These may be used later to help the visualisation of the netcdf data.
   if ( !this->isVisualise() )
   {
      out = ncDataReq;
//      ncDataReq("_VIEW") = view_.c_str();
//      ncDataReq("_VISUALISE") = out1;
//      ncDataReq("_CARTESIANVIEW") = viewReq;
//      ncDataReq("_ORIGINAL_REQUEST") = origReq_;
      return true;
   }

   // Create title information
   MvRequest titleReq("MTEXT");
   titleReq("TEXT_LINE_1") = this->title().c_str();
   titleReq("TEXT_FONT_SIZE") = 0.5;

   // Final output request
   out = viewReq + outd + titleReq;

   return true;
}

//
//  Title reflects data extraction type
//
string GribThermo::title()
{
   // Get first time information from the netcdf file
   MvNetCDF netcdf(this->ncFileName(),'r');
   MvNcVar *time = netcdf.getVariable(sTIME);
   NcDim* dim = time->getDimension(1);
   int size = dim->size();
   vector <char> vtime;
   time->setCurrent(0L);
   time->get(vtime,1,size);

   // Decode time info
   string str(vtime.begin(),vtime.end());
   string sdate = str.substr(0,8);
   string stime = str.substr(8,4);
   int istep    = atoi(str.substr(12,4).c_str());

   // Create title
   char  title[200];
   switch( pars_.pointType_ )
   {
      case eInterpolated:
         sprintf(title,"%s %s step %d [%.2f,%.2f] %s",
                 sdate.c_str(), stime.c_str(), istep,
                 pars_.lat_, pars_.lon_,
                 pars_.formulaPrintName().c_str() );
         break;

      case eGridPoint:
      {
         // Coordinates used may not be the same as the input ones
         MvNcAtt* tmpatt = netcdf.getAttribute("Coordinates");
         str = tmpatt->as_string(0);
         int ipos = str.find('/');
         float lat = atof(str.substr(0,ipos).c_str());
         float lon = atof(str.substr(ipos+1).c_str());
         sprintf(title,"%s %s step %d [nearest: %.2f,%.2f] %s",
                 sdate.c_str(), stime.c_str(), istep, lat, lon,
                 pars_.formulaPrintName().c_str() );
      }
      break;

      case eAreaAverage:
         sprintf(title,"%s %s step %d [%.1f,%.1f,%.1f,%.1f] %s",
                 sdate.c_str(), stime.c_str(), istep,
                 pars_.lat_, pars_.lon_, pars_.lat2_, pars_.lon2_,
                 pars_.formulaPrintName().c_str() );
      break;
   }

   return string(title);
}
