/***************************** LICENSE START ***********************************

 Copyright 2012 ECMWF and INPE. This software is distributed under the terms
 of the Apache License version 2.0. In applying this license, ECMWF does not
 waive the privileges and immunities granted to it by virtue of its status as
 an Intergovernmental Organization or submit itself to any jurisdiction.

 ***************************** LICENSE END *************************************/

//
// .NAME:
//  Presentable
//
// .AUTHOR:
//  Gilberto Camara, Baudoiun Raoult and Fernando Ii
//
// .SUMMARY:
//  Defines an abstract base class for the page hierarchy in PlotMod.
//  The Presentable class is an example of the Composite "pattern".
//
//  Its methods allow for the insertion and removal of nodes and
//  leaves on a tree structure, and also for the drawing method.
//
//  The building of the tree structure, from a description,
//  is done by the Builder class.
//
//  Other applications that use Presentable (such as Matching) are
//  implemented by the Visitor class.
//
// .DESCRIPTION:
//  Refer to the "Design Patterns" book, page 163 for more details
//
//
// .DESCENDENT:
//  Page, SubPage, DataObject, Text, Legend
//
// .RELATED:
//  PlotMod, Builder, Visitor, Device
//
// .ASCENDENT:
//
//
#ifndef Presentable_H
#define Presentable_H

#include <inc_stl.h>

#include "MvRequest.h"
#include "MvIconDataBase.h"
#include "DeviceData.h"
#include "Visitor.h"
#include "PlotModConst.h"
#include "PaperSize.h"
#include "Location.h"
#include "MatchingInfo.h"

class PmContext;
class Canvas;
class GraphicsEngine;
class DrawingPriority;
class PlotModView;
class Presentable;
class ObjectInfo;

typedef map <string, pair<int, int> > DrawPriorMap;

//typedef set<int> MvIntSet;
typedef map<int,string> MacroVisDefMap;
//typedef list<Canvas*> CanvasList;

enum { GETBYDATAUNIT,GETBYVISDEF};

class PresentableObserver {

	friend class Presentable;

	set <Presentable*> observed_;
	void Dead(Presentable*);

	// Override this one
	virtual void Notify(Presentable&) = 0;

public:

	void Observe(Presentable&);
	void Forget(Presentable&);

	void ForgetAll();

	virtual ~PresentableObserver();
};

class Presentable {

public:

	// Contructor
	// Builds a presentable without a request
	// (used for data objects)
	Presentable();

	// Builds a presentable from a Request
	Presentable( const MvRequest& inRequest);

	Presentable(const Presentable&);

	virtual Presentable* Clone() const { return new Presentable(*this); }

	// Destructor
	virtual ~Presentable();

	// Members Information

	// Returns a reference to the object's associated request
	virtual MvRequest& Request()
		{ return myRequest_; }

	// true : presentable is a window (interactive)
        // false: presentable is not a window
	virtual bool IsAWindow()
		{ return myParent_->IsAWindow(); }

	virtual bool HasDrawTask ();
	virtual void HasDrawTask( bool yesno )
		{ hasDrawTask_ = yesno; }

    // Draws the presentable
    virtual void Draw ();
    virtual void DrawProlog ();
    virtual void DrawPageHeader () {}
    virtual void DrawTrailer () {}
    virtual void DrawTrailer ( MvRequest& );
    virtual void DrawTask ();    // Ask the server to draw me
    virtual bool CheckDrawTask(); // Check if there is any task to be performed

	// Methods for Inserting and Removing Children
        // Inserts a new child in the tree hierachy
        // Fails if the object is a leaf
        virtual void Insert (Presentable* p);

        // Remove an existing child from the tree hierarchy
        // Fails if the object is a leaf
        virtual void Remove (Presentable* p);

	virtual void CreateReply ( MvRequest& ) {}

	// Dealing with Drop
	virtual void Drop ( PmContext& context );
	virtual void DropSimulate ( PmContext& context );

	// Dealing with observers
	void NotifyObservers();

#if 0
        // Recursively copies children and attachments to new SuperPage hierarchy
	virtual void DuplicateChildren(const Presentable&);


	// Close All Windows Associated With a Presentable
	virtual void Close();
#endif

	// Zoom methods
//	virtual void ZoomRequest ( const Location&  ) {}
	virtual void ZoomRequest ( int, const Location&  ) {}
	virtual void ZoomRequest ( int, const string&  ) {}
	virtual void ZoomRequest ( int, int  ) {}
	virtual void ZoomInfo(int&, int&) {cout << "Should not be here" << endl;}

    // Layer methods
    virtual bool UpdateLayerTransparency  ( int, int )  { return false; }
    virtual bool UpdateLayerVisibility    ( int, bool ) { return false; }
    virtual bool UpdateLayerStackingOrder ( ) { return false; }

	// Replace the geographical area
	//virtual void  ReplaceArea ( const Location&  ) {}

#if 0
	// Update the view based on a magnification
	virtual void  MagnifyRequest ( const Location&  ) {}

	virtual void  InputCoordinates ( const Location&, int ) {}

	virtual void  ExecuteToolPlusPointScript ( const Location& ) {}

	virtual void  ExecuteToolPlusAreaScript ( const Location& ) {}

	virtual void SetInputPresentableId ( int ) {}

	// Methods used in relation to Printing actions
	// Printing the presentable
	virtual void  Print();

	// Preview the printing
	virtual void  PrintPreview();

	// Methods in relation to User Preferences 
	virtual Preferences& GetPreferences()
		{ return myParent_->GetPreferences(); }

	// Setup the preferences according to the desired user setup
	virtual void  EditPrinterPreferences () {}

        // Edit the interface preferences
	virtual void  EditInterfacePreferences() {}


        // Edit the map preferences
	virtual void  EditMapPreferences() {}

        // Edit the coast preferences
	virtual void  EditCoastPreferences() {}

	// Methods associated with the callbacks on the user interface
	virtual void  SetCursorType ( CursorType ) {}
	virtual void  SetInputType  ( InputType ) {}

	virtual void  PostMessage ( const char*, const char* ) {}

	virtual void  PrintJob () {}  //D Remove later
	virtual void  PrintJob ( string& ) {}

	virtual void  Undo();
	virtual void  Redo();
	virtual void  Reset();

	virtual void  StartMapEditor() {}

	virtual void  SaveView() {}

	virtual void  OKCallback () {}

	virtual void  SelectProjection ( int, double  = 0.0 ) {}

	virtual void  MouseMode () {}

	virtual void ChangeFontType ( bool ) {}

	// Methods associated with  scrolling
	virtual bool CanGotoNext() { return false; }
	virtual bool CanGotoPrevious() { return false; }
	virtual bool CanScroll();

	virtual void GotoNext(bool  = false) {}
	virtual void GotoPrevious(bool  = false) {}

	virtual void GotoFirst() {}
	virtual void GotoLast() {}

	virtual bool CanChangeGeography() { return false; }

        virtual void ScrollByScreen(bool) {}

	// Dealing with Contents
	virtual Cached Name();
	virtual Cached Class();
	virtual Cached SuperPageName();

	virtual Cached Type() { return Cached("Presentable"); }

	// Generates a Request for all data contained in the presentable's branch
	virtual MvRequest  ContentsRequest ();

	// Visit my children and find their information
	virtual void ChildrenContents( MvRequest& request );
	
	// Include my own contents in the request
	virtual void IncludeContents ( MvRequest& request );

	// Show a contents visitor
	// Set or get the data unit from contents.
	virtual void ContentsDataUnit( int ) { }
        virtual int ContentsDataUnit() { return 0; }
#endif

    // Dealing with Contents
    virtual string MacroPlotIndexName(bool = false) { return ""; }
          //	{ return myParent_->MacroPlotIndexName(useIndex);}
    virtual void contentsRequest( MvRequest& ) {}

	virtual string Name();
	virtual string MacroName();

    // Flag indicanting if drop came from Contents Window
    virtual void DropShouldStay( bool ) { }
	virtual bool DropShouldStay() { return false; }

    // Register to draw trailer later
    virtual void RegisterDrawTrailer () {}

	// Remove from tree structure, not database
	virtual void RemoveIcon ( MvIcon& ) {}
	virtual void RemoveMyData ();
	virtual void RemoveAllData () {}
	virtual void RemoveAllDataChildren ();

#if 0
	virtual void RemoveMyVisDef (const MvRequest* req=0);
	virtual void RemoveMyVisDef (MvIconList&, const MvRequest* req=0);
	virtual void RemoveAllVisDef ();
	virtual void RemoveAllVisDefChildren ();
	virtual void RemoveMyPText ();
	virtual void RemoveAllPText ();
	virtual void RemoveAllPTextChildren ();
#endif

	// Dealing with Macro
	// Create Macro code from the Display Window
	virtual void GenerateMacro ( ) { }

	// Generate a description of the presentable
	virtual void DescribeYourself ( ObjectInfo&   ) {}

	// Describe the drops (data units and visdefs)
	virtual void DescribeDrops ( ObjectInfo& myDescription, MacroVisDefMap &vdMap );

#if 0
	// Save your contents to a request
	virtual void Save () {}

	virtual void SaveMap () {}

	// Save the contents of your children
	virtual void SaveChildren ();

        // Update a visdef request for all children. Overridden if needed.
        virtual void UpdateVisDef(MvRequest &) {}

        // Remove visdef ( or part of it ) for all children. Overridden if needed.
        virtual void RemoveVisDef(const char *) {}
#endif

	// Methods for Handling the Children and the Father
	const Presentable* Parent() const
		{ return myParent_; }

     void Parent( Presentable* pp ) 
          { myParent_ = pp; }

	// Return number of children
	int NrOfChildren() const { return childList_.size(); }

	// Update list of childeren in case of changes ( ie from contents).
//	virtual void UpdateChildList () {}

	// Methods which traverse the tree hierarchy
        // Finds a sub-branch of the tree hierarchy whose
        // top node has the same ID as "treeNodeId"
	Presentable* FindBranch (const int treeNodeId) const;

	// Finds a sub-branch of the tree hierarchy whose
	// top node has a View specified by "viewName"
	Presentable* FindBranch (string viewName);

	// Finds a sub-branch of the tree hierarchy whose
	// top node has a request specified by "requestName"
  	void FindBranchesByRequest(string requestName,vector<Presentable*>& presVec);

        // Finds the superpage associated to a presentable
	virtual Presentable* FindSuperPage () const;

	// Methods used for Matching

	// Indicates that the presentable (or one of its
	// children) is on a tree branch with dataObjects
//	virtual void DataInserted();

	// Indicates that the presentable contains no dataObjects
//	virtual void DataRemoved();

	// Informs if the presentable has a dataObject
	virtual bool HasData();

	// Informs if the presentable has a child with a dataObject
	virtual bool ChildHasData();

         // Indicates that a presentable is not matched
	virtual void NotMatched()
		{ hasMatch_ = false; }

	// Indicates that a match has been found for the presentable
	virtual void MatchFound()
		{ hasMatch_ = true; }

	// Retrives the information about the object's matching
	virtual bool HasBeenMatched()
		{ return hasMatch_; }

	// Initializes all of the presentables' children 
	// to be un-matched (to be used in the beginning of
	// the matching procedure)
	virtual void InitMatching();

	// Indicates a reference matching info for the presentable
	virtual void SetMatchingInfo(const MatchingInfo& dataInfo)
		{ matchingInfo_ = dataInfo; }

	// Retrive the matching info associated with the treeNode
	virtual const MatchingInfo& GetMatchingInfo() const
		{ return matchingInfo_; }

	// Methods used for Drawing
       // Draws the data
//        virtual void DrawData () {}

    // Draws the presentable's children
    virtual void DrawChildren ();

    // Store the Layer info
    virtual void DrawLayerInfo ( const int );

#if 0
	// Draw the background (ask my view to do it )
	virtual void DrawBackground ( Canvas& , bool = false ) {}

	// Draw the foreground (ask my view to do it )
	virtual void DrawForeground ( Canvas& , bool = false ) {}

//D	virtual void DrawFrame ( Canvas&  ) {}
	// Draw the frame (ask my view to do it )
	//  

	// Draw the frame (ask my view to do it )
	virtual void DrawNewPage ( Canvas&  ) {}
#endif

	virtual void DrawText ( ) {}

	virtual int PageIndex();

	virtual int PaperPageIndex() const;
	virtual void PaperPageIndex ( int );

	virtual bool PrintAll () { return false; } 
	virtual void PrintAll ( bool ) {} 

	virtual int PaperPageNumber ( int ) { return -1; }

        virtual void DrawDataVisDef() {}

	virtual void EraseDraw ( int  = 0 ) {}
	virtual void EraseDefaultDraw ();

//        virtual void RetrieveTitleFromMagics( MvIcon& ) {}		

	// Dealing with DrawingArea, return from Page or EditMapPage
//	virtual Location GetDrawingArea ();
//	virtual void SetDrawingArea ( Location );

	// Methods which access the MvIconDataBase
        // This method is used to retrieve a valid visdef.
	// It is called when there are no visual definitions
	// associated to a data unit. If last arg is GETBYDATAUNIT,
        // the first arg is interpreted as a dataunit class.
        // If last arg is GETBYVISDEF, first arg is the visdef class to retrive.
        virtual bool DefaultVisDefList (const char* className, 
					MvIconList& visdefList,
					int type = GETBYDATAUNIT);

        // Overridden by Page/SubPage
        virtual bool CheckValidVisDef(MvIcon &,MvRequest&)
		{ return true; }

   // Retrieve the Text list associated to the node
   virtual bool RetrieveTextList ( MvIconList& );

   // Retrieve the Legend associated to the tree
   virtual bool RetrieveLegend ( MvIcon& );

	// Retrieves the icon database associated to the
	// tree branch (normally, a superpage)
	virtual MvIconDataBase& IconDataBase() const
		{ return myParent_->IconDataBase(); }

	//  Dealing with Devices and GraphicsEngine
//	virtual Device&          GetDevice () const ;
	virtual GraphicsEngine& GetGraphicsEngine() const;
	virtual PlotModView& GetView() const;
//	virtual bool FindView(string viewName)
//		{ return false; }

	// Dealing with Canvas
	virtual Canvas& GetCanvas() const ;
	virtual void    SetCanvas ( Canvas * ) {}
	virtual bool    HasCanvas()
		{ return false; }
//	virtual CanvasList GetCanvases ();

	// Dealing with Magnify, to be override in child class MagnifyPage
//	virtual Canvas* GetOwnerCanvas()
//		{ return 0; }

//	virtual int     MagFactor ()
//		{ return 1; }

	// Dealing with Printing
	virtual void PendingDrawingsAdd();
	virtual void PendingDrawingsRemove();
	virtual bool PendingDrawings()
		{ return  (  pendingDrawings_ != 0 ); }


//	virtual bool DrawingIsFinished();
//		{ return  (  pendingDrawings_ == 0 ) && 
//			  (  hasNewpage_  == true  ); }

	virtual void HasReceivedNewpage( bool flag)
	     { hasNewpage_ = flag; }

	virtual bool HasReceivedNewpage()
	     { return hasNewpage_; }

// Save title information
//	virtual void SaveTitle ( MvRequest& ) {}

	// Dealing with DeviceData
	DeviceData* ReleaseDeviceData ( );
	void        SetDeviceData ( DeviceData* devdata );
	DeviceData* GetDeviceData() const
		{ return deviceData_.get(); }

	virtual void Visit ( Visitor& v )
		{ VisitChildren ( v ); }

	void VisitChildren ( Visitor& v );

	// Set/Get the location for the presentable
        virtual void SetLocation ( Location pos )
		{ myLocation_ = pos; }

	Rectangle GetLocation ( )
		{ return myLocation_; }

//	virtual void UpdateWidget ( MvRequest& , const int, int& ) {}

	// Members and Object Information
//	virtual bool IsSuperPage()
//		{ return (myRequest_.getVerb() == PLOTSUPERPAGE); }

	// true : presentable is visible and will be drawn
        // false: presentable is not visible
	virtual bool IsVisible()
		{ return isVisible_; }

	// Set the visibility for the presentable
	virtual void SetVisibility ( bool visibility )
		{ isVisible_ = visibility; }
	
	// Returns the object's unique identification
	int Id() const { return presentableId_ ; }

	virtual long PictureId(const char *, bool = false)
		{ return 0; }

	virtual PaperSize GetChildSize ( Presentable& );

	virtual PaperSize GetMySize()
		{ return myParent_->GetChildSize ( *this ); }

        virtual void NeedsRedrawing(bool yesno)
		{ needsRedrawing_ = yesno; }

        virtual bool NeedsRedrawing()
		{ return  needsRedrawing_; }

#if 0
	virtual ZoomStacks* GetZoomStacks()
		{ return 0; }

	virtual void SetWidgetSensitive ( const char*, bool ) {}

	void  InAnimation ( bool inAnimation )
		{ inAnimation_ = inAnimation; }

	bool InAnimation ()
		{ return inAnimation_; }

	// Update SuperPage's size
	virtual void SetMySize ( double, double ) {}

	// Update Page's size
	virtual void SetPageSize ( double, double, double, double ) {}
#endif

	// Set the projection associated to the canvas
	virtual void SetProjection ( const MvRequest& ) {}

//	virtual bool HasProjection()
//		{ return false; }

	virtual bool CanDoZoom()
		{ return false; }

//	virtual PmProjection* GetProjection()
//		{ return 0; }

	virtual void SetDrawPriority( MvRequest& ) {}

	virtual DrawingPriority& GetDrawPriority ();

	// Indicate//determine whether a page is visited in background
	virtual bool VisitInBackground() { return false; }
	virtual void VisitInBackground(bool ) {}

	virtual bool RemoveData()
		{ return removeData_; }

	virtual void RemoveData( bool yesno )
		{ removeData_ = yesno; }

	// Used to indicate Contents drop on a window
	bool DroppedHere () { return droppedHere_; }
	void DroppedHere ( int Id ) { droppedHere_ = ( Id == presentableId_ ); }

	virtual void InsertPText ( MvRequest& textRequest );
	virtual void InsertLegend ( const MvRequest& legRequest );
	virtual Page* InsertOnePage ( MvRequest& )
		{ cout << "Should not be here" << endl; return 0; }

//U	virtual void EraseForeDraw () {}

        // Sets redrawing flags and does AddTask if needed
        virtual void RedrawIfWindow(int = 0) {}

//	virtual void SetBackgroundDraw ( bool  ) {}

//	virtual bool GetBackgroundDraw ()
//		{ return 0; }

//	virtual int GetProjectionId ()
//		{ return 0; }

	virtual void DataUnit(MvIcon&) {} 

   // Dealing with Export plotting
   virtual bool ExportPlot ( MvRequest*, bool = false ) { return false; }

	// Dealing with Print plotting
	virtual void PrintFile ( MvRequest& ) { }

	virtual void SetRequest(const MvRequest & );

	// Dealing with rebuilding the visualization tree
	void Refresh ( bool refresh ) { refresh_ = refresh; }
	bool Refresh () { return refresh_; }

protected:

        bool DescribeVisDefList( ObjectInfo &myDescription,
				 MvIconList& visdefList,MvIcon &du,
				 Cached &listDrops, bool addToDrops,
				 MacroVisDefMap & vdMap);

	// Members
	static int treeNodeCounter_;	// used to obtain unique Id
	MvRequest myRequest_;		// request used to build me
	int presentableId_;			// unique Id
	Presentable* myParent_;		// next level up on the tree
	auto_ptr <DeviceData > deviceData_;	// information on the device
	Location myLocation_;       	// position of the presentable

	// Definition for the tree hierarchy of the Presentable class
	typedef list <Presentable* > MvChildList;
	typedef list <Presentable* >::iterator MvChildIterator;
	typedef list <Presentable* >::const_iterator MvChildConstIterator;

	MvChildList childList_;		// list of children
	bool hasMatch_;			// indicates that a object has been matched
	MatchingInfo matchingInfo_;	// matching information

	bool isVisible_;			// indicates if the presentable is visible

	bool hasNewpage_;			// indicates if the presentable
							// has received a newpage directive

	int pendingDrawings_;		// number of pending drawings
	bool needsRedrawing_;		// Needed when scrolling to find out 
							// if presentable needs redrawing

//        bool                    inAnimation_;

	bool removeData_;			// To be used when removing icons
							// True if data is to be removed
							// False removes data drawing only

	bool hasDrawTask_;			// Used to indicate if the presentable
							// has been put in the Task list to be drawn

	bool droppedHere_;			// indicates if Contents Drop was
							// done in this window

	bool refresh_;				// tag to rebuild/not the visualization tree

private:

	// No copy allowed
	Presentable& operator=(const Presentable&);

	// Observer
	set<PresentableObserver*> observers_;
//	bool dying_;

	friend class PresentableObserver;

	void Add(PresentableObserver*);
	void Remove(PresentableObserver*);
};

#endif
